"""
This module contains procedures for exploring and downloading data on
ARM Data Discovery.

"""

from .get_armfiles import download_data
from .get_asos import get_asos
from .get_cropscape import croptype
